;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "auto-package-update" "20260601.1804"
  "Automatically update Emacs packages."
  '((emacs "24.4")
    (dash  "2.1.0"))
  :url "https://github.com/rranelli/auto-package-update.el"
  :commit "e966c6c95de1742d867250dc15b1c6bd570b6ea5"
  :revdesc "e966c6c95de1"
  :keywords '("package" "update"))
